# snippets from: http://docs.python.org/library/sqlite3.html#module-sqlite3

import sqlite3 as sqlite
#from pdb import set_trace
import ctypes as C
from struct import *

from reportlab.pdfgen.canvas import Canvas
from reportlab.lib.pagesizes import letter
from reportlab.lib.pagesizes import A4
from reportlab.lib.units import cm, mm, inch, pica

from pyPdf import PdfFileWriter, PdfFileReader

import wx

version = 2

class SqLiteTests():
    """
    Do some testing woth sqlite3
    """

    def setUp(self, name):
        # Make connection to database
        self.cx = sqlite.connect(name)

    def tearDown(self):
        # Close the database connection
        self.cx.close()

    def draw_annotation_pdf(self, blob, pdf):
        pdf.setStrokeColorRGB(0, 0, 0)
        pdf.setFillColorRGB(0, 0, 0)

        # round line joints and endings
        pdf.setLineJoin(1)
        pdf.setLineCap(1)
       
        # ScbPageAttributes
        #print "ScbPageAttributes:"
        #- minor_version
        i = 0
        s = 4
        minor_version = unpack('l', blob[i:i+s])[0]
        #print "-minor_version: " + str(minor_version)
        i += s

        #- style ScbPageStyle
        #print "- style ScbPageStyle:"
        # |- orientation
        s = 4
        orientation = unpack('l', blob[i:i+s])[0]
        #print " |-orientation: " + str(orientation)
        i += s
        
        # |- bgColor ScbColor
        #   |- pixel
        #   |- red
        #   |- green
        #   |- blue
        s = 12
        bgColor = unpack('lhhh0l', blob[i:i+s])
        #print " |-bgColor (ScbColor): " + str(bgColor)
        i += s
        
        #- area ScbRect
        # |- left
        # |- top
        # |- right
        # |- bottom
        s = 16
        area = unpack('llll', blob[i:i+s])
        #print "-area (ScbRect): " + str(area)
        i += s
        
        #- stokes_number
        s = 4
        stokes_number = unpack('l', blob[i:i+s])[0]
        #print " |-stokes_number: " + str(stokes_number)
        i += s

        #max values
        max_x = 0
        max_y = 0
        min_x = 9999
        min_y = 9999
        
        #- Array of stokes (with ScbStrokeAttributes and array of ScbPoints)
        #print "- array of stokes:"
        for stroke in range (stokes_number):
            #print "    [%3d]" % stroke
            # |- ScbStrokeAttributes attributes
            #    0:color, 1:zoom, 2:layer, 3:shape_id, 4:size_id, 5:points_number
            s = 24
            attributes = unpack('lfllll', blob[i:i+s])
            #print "     |-attributes: " + str(attributes)
            color = attributes[0]
            zoom = attributes[1]
            size_id = attributes[4]
            points_number = attributes[5]
            i += s

            # Adapt the line width
            pdf.setLineWidth(size_id/3.0)

            #- Array of points
            #print "     |-array of stokes (%d):" % points_number
            first = 1
            points = []
            path = pdf.beginPath()
            for point in range (points_number):
                #print "        [%3d]" % point,
                #    - point
                s = 12
                #point =  unpack('lll', blob[i:i+s])
                #print point
                x, y, pressure =  unpack('lll', blob[i:i+s])
                #print "%6.1f %6.1f %d" % (x, y, pressure),
                # input device is 160DPI, so 1280/160=8, 1024/160=6.4, 768/160=4.8
                #                            1280/72=17.8 1024/72=14.2, 768/72=10.7
                x = ((x * 100.0 / zoom)) / (160.0/72.0*(1-0.0625))
                y = ((y * 100.0 / zoom)) / (160.0/72.0*(1-0.0625))
                # adjust from area size to window size
                #x = ((x * 100.0 / zoom)) * (winWidth/area[2])
                #y = ((y * 100.0 / zoom)) * (winHeight/area[3])
                #print "%6.1f %6.1f %d" % (x, y, pressure),
                # determine extreme values
                if x > max_x:
                    max_x = x
                if y > max_y:
                    max_y = y
                if x < min_x:
                    min_x = x
                if y < min_y:
                    min_y = y
                i += s
                
                if first == 1:
                    first = 0
                    path.moveTo(x, y)
                else:
                    path.lineTo(x, y)
                    
            #print ""
            if points_number > 1:
                pdf.drawPath(path, True, False)
        
        #print blob[0:16].encode('hex')
        #print blob[16:32].encode('hex')
        #print blob[32:48].encode('hex')
        print "Used Window (%f, %f, %f, %f)" % (min_x, min_y, max_x, max_y)

    def has_page(self, pages, fileNr, pageNr):
        # find pages with annotations
        for fNr, pNr, blob in pages:
            if fileNr == fNr:
                if pageNr == pNr:
                    return True
        return False

    def merge(self, directory, mode, logger):
        global app
        logger.AppendText("Merge\n")

        # list of all pages, in order of annotation
        pages  = []
        # list of all files
        files = []
        fileNrs = []
        
        self.cu = self.cx.cursor()
        
        # use file_metadata to extract name (match via file_id)
        self.cu.execute('SELECT * FROM file_metadata')
        allentries=self.cu.fetchall()
        for x in allentries:
            print x
            files.append((x[0], x[1]))
            #print "Item: " + str(x[0]) + " " + str(x[1]) + " " + str(x[2]) + " " + str(x[3])
            

        # use annotations to get scribbles
        self.cu.execute('SELECT * FROM annotations')
        entries = self.cu.fetchall()
        index = 0
        
        for b in entries:
            logger.AppendText( "Type: %s FileNr: %d, page: %d\n" % (b[1], b[2], b[4]))
            if b[1] == 'scribble':
                # check type: only add 'scribbles'!!
                fileNr = b[2]
                if fileNr not in fileNrs:
                    fileNrs.append(fileNr)
                pageNr = b[4]
                blob = b[8]
                pages.append((fileNr, pageNr, blob))
        logger.AppendText( "Files : %s\n" % str(files))
        #logger.AppendText( "Pages : %s\n" % str(pages))

        # Create pdf files with the annotations
        for fileNr in fileNrs:
            # find name
            for nr, fileName in files:
                if nr == fileNr:
                    break
            # use filename from db
            name = "%s\\Anno_%s" % (directory, fileName)
            # determine size
            name_in   = "%s\\%s" % (directory, fileName)
            input2 = PdfFileReader(file(name_in, "rb"))
            # do not use the first page, it can have a different size
            if input2.getNumPages() > 4:
                pageNr = 4
            else:
                pageNr = input2.getNumPages()-1
            page = input2.getPage(pageNr)
            width = page.mediaBox.getWidth()
            height = page.mediaBox.getHeight()
            logger.AppendText( 'Page %d size %d x %d\n' %(pageNr, width, height))
            
            logger.AppendText( 'Generating: %s\n' % name)
            if mode == 1:
                pdf = Canvas(name, pagesize=(width, height), bottomup = 0)

            # find pages with annotations
            for fNr, pageNr, blob in pages:
                if fileNr == fNr:
                    logger.AppendText( "File %d has page %d\n" % (fileNr, pageNr))
                    if mode == 1:
                        self.draw_annotation_pdf(blob, pdf)
                        # finish page
                        pdf.showPage()
            if mode == 1:
                # and save
                pdf.save()

            # keep event processing alive
            while app.Pending():
                app.Dispatch()

        # next step is to merge new file with existing
        if mode == 1:
            for fileNr in fileNrs:
                # find name
                for nr, fileName in files:
                    if nr == fileNr:
                        break
                # use filename from db
                name_in   = "%s\\%s" % (directory, fileName)
                name_anno = "%s\\Anno_%s" % (directory, fileName)
                name_out  = "%s\\Gen_%s" % (directory, fileName)
                logger.AppendText( 'Generating: %s\n' % name_out)

                # open the input and output files
                output = PdfFileWriter()
                input1 = PdfFileReader(file(name_anno, "rb"))
                input2 = PdfFileReader(file(name_in, "rb"))

                logger.AppendText( "%s has %d pages.\n" % (name_in, input2.getNumPages()))
                annoPageNr = 0
                for pageNr in range(input2.getNumPages()):
                    page = input2.getPage(pageNr)
                    logger.AppendText( 'Copying page %d size %d x %d\n' %(pageNr+1, page.mediaBox.getWidth(), page.mediaBox.getHeight()))
                    if self.has_page(pages, fileNr, pageNr+1):
                        page2 = input1.getPage(annoPageNr)
                        logger.AppendText( '  Merging Page %d size %d x %d\n' %(annoPageNr+1, page2.mediaBox.getWidth(), page2.mediaBox.getHeight()))
                        annoPageNr += 1
                        page.mergePage(page2)
                    output.addPage(page)
                    # keep event processing alive
                    while app.Pending():
                        app.Dispatch()

                # finally, write "output" to document-output.pdf
                outputStream = file(name_out, "wb")
                output.write(outputStream)
                outputStream.close()
     
            logger.AppendText("End Merge\n\n")

#-------------------------------------------------------------------
#                   graphics stuff
class mergeFrame(wx.Frame):
    """Frame for application"""
    def __init__(self, parent, title):
        wx.Frame.__init__(self, parent, title=title, size=(400,200))
        self.CreateStatusBar()

        # Create a log widget
        self.logger = wx.TextCtrl(self, pos=(300,20), size=(200,300), style=wx.TE_MULTILINE | wx.TE_READONLY)
        self.logger.AppendText("mergeTool version:%d" % version)

        # Setting up the menu
        filemenu = wx.Menu()
        menuOpen = filemenu.Append(wx.ID_OPEN, "&Open"," Pick directory to process")
        menuAbout = filemenu.Append(wx.ID_ABOUT, "&About", " Information about this program")
        filemenu.AppendSeparator()
        menuExit = filemenu.Append(wx.ID_EXIT,"E&xit"," Terminate the program")

        # Creating the menubar.
        menuBar = wx.MenuBar()
        menuBar.Append(filemenu,"&File") # Adding the "filemenu" to the MenuBar
        self.SetMenuBar(menuBar)  # Adding the MenuBar to the Frame content.

        # Set events
        self.Bind(wx.EVT_MENU, self.OnOpen, menuOpen)
        self.Bind(wx.EVT_MENU, self.OnAbout, menuAbout)
        self.Bind(wx.EVT_MENU, self.OnExit, menuExit)

        self.Show(True)
        
    def OnAbout(self,e):
        # A message dialog box with an OK button. wx.OK is a standard ID in wxWidgets.
        dlg = wx.MessageDialog( self, "Merges annotations with PDF files", "mergeTool", wx.OK)
        dlg.ShowModal() # Show it
        dlg.Destroy() # finally destroy it when finished.

    def OnExit(self,e):
        self.Close(True)  # Close the frame.
        
    def OnOpen(self,e):
        """ Pick a directory"""
        self.dirname = ''
        dlg = wx.DirDialog(self, "Choose a directory", self.dirname)
        if dlg.ShowModal() == wx.ID_OK:
            self.dirname = dlg.GetPath()
            # do the work here
            sd = SqLiteTests()
            db_name = "%s\\metadata.db" % self.dirname
            sd.setUp(db_name)

            sd.merge(self.dirname, 1, self.logger)
            
            sd.tearDown()
        dlg.Destroy()

#-------------------------------------------------------------------
#                   main

app = wx.App(False)
frame = mergeFrame(None, "mergeTool")
app.MainLoop()
    
    
